from mpi4py import MPI
from copy import deepcopy
import numpy as np
from math import floor
from itertools import groupby
from .tmatrix import make_dumb_tmatrix
import time
from . import metropolis
from .make_regr import make_regr
from susmost.savexyz import save_cells_as_xyz

"""
m.params_log:

0)	additive_params[0] / lattice.cells_count
1)	additive_params[1] / lattice.cells_count
........
-8)	in-site avgE / lattice.cells_count
-7)	in-site entropy / lattice.cells_count
-6)	in-site min_E / lattice.cells_count
-5)	in-site free_E / lattice.cells_count
-4)	acceptance rate
-3)	curE.internal / lattice.cells_count
-2)	curE.sum() / lattice.cells_count
-1)	T_idx
"""

def make_additive_params(states, param_names):
	return [[[props[pname].value for pname in param_names] for _,props in s.EP_list()] for s in states]
	
def make_metropolis(lat_task, lattice, T, k_B, param_names=None, precoverage = None, seed=None):
	"""
		Create Metropolis object for specified LatticeTask and ReGraph objects
		
		Parameters:
			:lat_task: LatticeTask object
			:lattice: sizes of the lattice in unit cells. Must be either a number, a 2-element list or a 3-element list, all variants are transformed into 3-element list with the following equivalence: ``n == [n,n] == [n,n,1]``
			:T: list of temperatures in Kalvin units for replica exchange (*aka* parallel tempering), number of temperatures must be equal to number of MPI processes
			:k_B: Boltzmann constant, used as energy scale parameter
			:param_names: list of names of states properties that will be sampled in Metropolis iterations as simulaion parameters; default value - ['coverage']
			:precoverage: index of the state that will be used for initialization of the lattice, by default the state with zero ``coverage`` parameter is used
			:seed: random number generator for Metropolis simulaion
		
		Returns:
			Metropolis object
	"""
	
	regr = make_regr(lattice, lat_task)

	# key constants:
	cells_count	=	len(regr)
	edges_count	=	lat_task.edges_count
	lc_count 	= 	lat_task.states_count
	#----------------------------------
	try:
		T_list = [t for t in T]
	except TypeError:
		T_list = [T]
	#----------------------------------
	if param_names is None:
		param_names = ['coverage']
	#----------------------------------
	if seed is None:
		seed = int(time.time())
	#----------------------------------
	if precoverage is None:
		precoverage = lat_task.zero_coverage_state_index
	
	precov_state_index = int(precoverage)
	cells = np.full(cells_count, precov_state_index, dtype=int)	# fill by empty_state_index

	#----------------------------------

	
	allowed_cells = np.full( (cells_count,lc_count), True, dtype=bool)	# all allowed
	transitions = make_dumb_tmatrix(lat_task.states) # any to any allowed
	internal_energies = [[E for E,_ in s.EP_list()] for s in lat_task.states]	# from adsorption energies, s.E - is a  list of
	lattice_graph=regr.as_nparray(edges_count) # fill lattice_graph from regr
	interaction = lat_task.IM.asarray()  # fill interaction from IM
	additive_params = make_additive_params(lat_task.states, param_names)
	
	m = metropolis.Metropolis(lattice_graph, cells, allowed_cells, interaction, internal_energies, transitions, T_list, additive_params, k_B, seed) 

	m.phys_diffusion = False
	m.E_inf = lat_task.INF_E
	m.PT_period = 10 * len(cells) 	# Each m.PT_period steps we try to switch the temperature
	m.diff_cui_min = 1 # Minimal
	m.diff_cui_max = 1
	
	m.regr = regr
	m.lattice_task = lat_task
	m.param_names = param_names
	
	return m

def run(m, log_periods_cnt, log_period_steps=None, log_callback=None, params_period_steps=None, relaxation_steps = None, bcast_full_params_log = True, traj_fns=None):
	"""
		Perform Metropolis iterations
		
		Parameters:
			:m: Metropolis object
			:log_periods_cnt: number of log periods to perform
			:log_period_steps: number of Metropolis iterations in a single log period. By default ``m.cells_count`` value is used so log period is equivalent to conventional *Monte Carlo step* notion
			:log_callback: function to be called after each log period, ``log_callback`` function must take ``m`` as a parameter and can return any object as result. List of all objects returned from ``log_callback`` will be returned as a result of overall ``run()`` call. By default is None, so nothing is logged.
			:params_period_steps: period for sampling of energy and state properties (see notes on ``make_metropolis()`` function). Samples are stored in ``full_params_log`` attribute of ``m`` object. Meassured in Metropolis iterations. Default value ``log_period_steps//30``.
			:relaxation_steps: number of Metropolis iterations to be skipped before logging and parameters sampling. Default value ``log_periods_cnt * log_period_steps``
			:bcast_full_params_log: set to False if parameters samples should not be broadcasted to all MPI processes. By default ``True``, so every processes receives complete sample of parameters.
			:traj_fns: list of XYZ file names to be used to save snapshots of the lattice each ``log_period_steps`` iteration. ``traj_fns[i]`` is used to save samples from *i*-th replica (see notes on ``make_metropolis()`` function).
			
		Results:
			List of objects returned by ``log_callback`` function calls
			
			After the function call ``m.full_params_log`` contains three-dimensional array with samples of simulation parameters. Parameter values are averaged over lattice sites.
			
			``m.full_params_log[i,j,k]`` is the value of the *k*-th parameter on *j*-th time period from *i*-th replica. Parameter index ``k`` correspond to the *k*-th item of ``param_names`` argument of ``make_metropolis()`` function. In addition to state parameters, several system-wide parameters are included into ``m.full_params_log``. In particular:
			
				- ``m.full_params_log[i,j,-1]`` index of the temperature of this sample
				- ``m.full_params_log[i,j,-2]`` total energy of the system per one lattice cell
				- ``m.full_params_log[i,j,-3]`` lateral interactions energy (total energy minus adsoprtion energy) per lattice cell
				- ``m.full_params_log[i,j,-4]`` acceptance rate of Metropolis iterations
				
				- ``m.full_params_log[i,j,-5]`` free adsorption energy per lattice cell
				- ``m.full_params_log[i,j,-6]`` ground state (minimal) adsorption energy per lattice cell
				- ``m.full_params_log[i,j,-7]`` entropy of adsoprtion complexes per lattice cell
				- ``m.full_params_log[i,j,-8]`` mean internal energy (enthalpy) of adsoprtion complexes per lattice cell
				
				- ``m.full_params_log[i,j,:-8]`` samples of states parameters - see ``param_names`` argument of ``make_metropolis()`` function.
	"""

	if log_period_steps is None:
		log_period_steps = m.cells_count

	if relaxation_steps is None:
		relaxation_steps = log_periods_cnt * log_period_steps

	if params_period_steps is None:
		params_period_steps = log_period_steps//30
	
	if log_callback is None:
		def log_callback(m):
			return None
	 
	if traj_fns is not None:
		assert len(traj_fns) == len(m.T_list), "Not enough trajectory filenames: {} instead of {}".format(len(traj_fns), len(m.T_list))
		def write_traj(m):
			save_cells_as_xyz(traj_fns[m.T_idx], m, comment='E')
	else:
		def write_traj(m):
			pass
		

	result = {}  # dict of temperatures to lists with log_callback() results  
	m.params_log_period = log_period_steps
	m.run(relaxation_steps)
	# assert all([pl.enabled for pl in m.param_limits]), [pl.enabled for pl in m.param_limits]
	m.full_params_log = []
	m.params_log_period = params_period_steps
	for i in range(log_periods_cnt):
		m.run(log_period_steps)
		m.full_params_log += [deepcopy(m.params_log)]
		res_i = log_callback(m)
		result.setdefault(m.T_idx,list()).append(res_i)
		write_traj(m)
	#print "Gathering callback results..."
	result = MPI.COMM_WORLD.gather(result,root=0)
	#print "Gathering callback results... done!"

	if MPI.COMM_WORLD.Get_rank() == 0:
		aggr = result[0]  # use root dict as initial for aggregated dict
		for dict_i in result[1:]:
			for T_idx, res_log in dict_i.items():
				aggr.setdefault(T_idx, list()).extend(res_log)
				
		result = [res_log for T_idx, res_log in aggr.items()]
		
		assert all([len(log_i) == log_periods_cnt for log_i in result]) # check for each temperature
	
	result = MPI.COMM_WORLD.bcast(result,root=0)

	full_params_log = np.concatenate(m.full_params_log)	# [time, param]
	full_params_log = MPI.COMM_WORLD.gather(full_params_log,root=0) # [rank, time, param]
	if MPI.COMM_WORLD.Get_rank() == 0:
		full_params_log = np.swapaxes(full_params_log, 0, 1)	# # [time, rank, param]
		
		# sort by the last element (T_idx)
		full_params_log = np.array([sorted(r, key=lambda v: v[-1]) for r in full_params_log]) # [time, T_idx, param]
		full_params_log = np.swapaxes(full_params_log, 0, 1)	# [T_idx, time, param]

	if bcast_full_params_log:
		#print "bcast full param log..."
		m.full_params_log = MPI.COMM_WORLD.bcast(full_params_log, root=0)
		#print "bcast full param log... done!"
	else:
		m.full_params_log = full_params_log


	return result
	
"""
def make_CDF(y):
	x = sorted(y)
	N = len(y)
	p = 1./N
	CDF = []
	cum_CDF = 0.
	for x in x:
		CDF.append(cum_CDF)
		cum_CDF += p
	assert abs(cum_CDF - 1.0) < 1e-6
	return np.array(x), np.array(CDF)
"""

def bins_count(brange, group_step, origin):
	""""""
	"""
		see divisibles() in metropolis.cpp for implementation explanation
	"""
	totalBins = int(floor((brange[1] - origin)/group_step))
	excludeBins = int(floor((brange[0] - origin)/group_step))
	return totalBins - excludeBins


def make_histo(y, brange, group_step, origin=0.0):
	N = bins_count(brange, group_step, origin)
	assert N > 0
	h = np.histogram(y, bins=N, range=brange, density = False)
	return h[0] * (1./len(y)), h[1]
	
def group_by_column(data, grouping_column, brange, group_step, origin=0.0):
	""""""
	"""
		data[time, param]
		
		Returns:
			data[group, time, param]
	"""
	N = bins_count(brange, group_step, origin)
	assert N > 0
	def bin_index(r):
		idx = int(floor((r[grouping_column] - brange[0])/group_step))
		if idx < 0:
			return -1	# too small outlier
		if idx >= N:
			return N	# too big outlier
		return idx

	d =dict([ (k,list(g)) for k,g in groupby(sorted(data, key=bin_index), bin_index)])
	
	return [d.get(i, []) for i in range(N)]
	

def group_statistics(params_log, param_idx, lower_limit, upper_limit, step):
	""""""
	"""
		params_log[time, param_idx]
		Returns:
			means: [group, param], zeros for empty groups
			stds: [group, param], zeros for empty groups
			probs: [group]
			outliers_part: float from 0 to 1
	"""
	m = params_log.shape[1] # number of parameters
	
	#hist,bins = mc.make_histo(params_log[:, limit_param_idx], (lower_limit, upper_limit), limit_step)
	groups = group_by_column(params_log,  param_idx, (lower_limit, upper_limit), step)
	means = np.array([np.mean(g, axis=0) if len(g) > 0 else [0.]*m for g in groups]) # [group, param]
	stds  = np.array([np.std(g, axis=0)  if len(g) > 0 else [0.]*m for g in groups]) # [group, param]
	lens = np.array([len(g) for g in groups], dtype=float) # [group, param]
	probs = lens / sum(lens) if sum(lens) > 0 else [0.] * len(lens)
	outliers_part = 1.0 - sum(lens) / len(params_log)
	return means, stds, probs, outliers_part

		

	
class ACFAnalysis:
	def __init__(self):
		pass
	
	def __repr__(self):
		return "AC times: zero {0.zero_time:.3f}; exp {0.ac_time_exp:.3f}; int {0.ac_time_int:.3f} len: {1}".format(self, len(self.acf))

def autocorr_analysis(acf, m=10):
	r = ACFAnalysis()
	r.acf = acf
	r.zero_time = np.argmax(acf <= 0.)
	if r.zero_time == 0 :
		r.zero_time = len(acf)
		
	if r.zero_time > 2:
		x = np.arange(r.zero_time-1)
		y = np.log(acf[:r.zero_time-1])
		r.ac_time_exp = -sum(x*x)/sum(x*y)	# exponential AC time - least squares fitting to exponential shape of ACF
		
		r.ac_time_exp_rmse = sum(np.exp( x * (-1./ r.ac_time_exp) - acf[:r.zero_time-1])**2)**0.5 / (r.zero_time-1)
	else:
		r.ac_time_exp = 0.

	r.expected_acf_sputter = 1./(len(acf)**0.5)
	r.acf_sputter = [np.std(acf[i*len(acf)//m:(i+1)*len(acf)//m]) for i in range(m)]
	r.acf_mean = [np.mean(acf[i*len(acf)//m:(i+1)*len(acf)//m]) for i in range(m)]
	
	r.ac_time_int = sum(acf[1:r.zero_time])*2. + 1.	# integrated AC time
	return r

	
def autocorr_func(y_, m = None, method="fft"):
	""""""
	
	"""
		m - number of items lefter and righter of origin in autocorr_func
		returns 2*m+1 element array
		
		Note: 
		http://golem.fjfi.cvut.cz/wiki/Library/others/BenczeA_Autocorrelation_05.pdf
			Autocorrelation analysis and statistical consideration for the determination of velocity fluctuations in fusion plasmas
			DOI: 10.1063/1.1909200
			"Our result may be surprising as the relative scatter does not depend on the ns=Ns/DT event rate. 
			One would expect that  for  higher  event  rates  the  statistics  should  improve. However, in our
			 case the scatter of the correlation function is produced  by  random  overlapping  of  events;  
			 this  is  proportional  to nS.  The  mean  of  the  correlation  function  is  also proportional to
			 nS, therefore the relative scatter will not de-pend  on  the  event  rate.  On  the  other  hand,  
			 when DT is  increased  at  a  fixed  event  rate  the NS number  of  events  increases  as  well  
			 without  changing  the  random  coincidence between  the  events,  and  as  a  result  the  relative 
			 scatter  de-creases.  In  this  sense  the  number  of  measured  events  does improve statistics."
	"""
	N = len(y_)
	if m is None:
		m = N-1
	assert m < len(y_), ["ACF length must be smaller than sample size", m, N]
	y = y_ - np.mean(y_)
	std_y = np.std(y)
	assert (std_y != 0), ["Standart deviation of time series is zero. Can't compute autocorrelation", std_y]
	y /= std_y
	if method=="naive":
		c = np.correlate(y, y, 'full')
		assert len(c) == (2*N-1), [len(c), N]
		Ns = np.array(list(range(1,N)) + [N] + list(range(N-1,0,-1)))
		c = c / Ns
		c =  c[len(c)/2 :]
	elif method=="fft":
		r = np.fft.rfft(y)
		s = abs(r)**2
		c = np.fft.irfft(s)
		c /= N
	else:
		assert False, method
	
	return c[: m ]
	

def stat_digest(m):
	if MPI.COMM_WORLD.Get_rank() != 0:
		return

	param_names = m.param_names + ["Energy"]
	param_indices = list(range(0, len(m.param_names))) + [-2]
	
	
	
	for Ti,T in enumerate(m.T_list):
		print ("Temperature # {} = {}".format(Ti,T))
		for pi, pn in zip(param_indices, param_names):
			print ("\tParameter # {} : '{}'".format(pi,pn))
			data = m.full_params_log[Ti,:,pi]
			mean = np.mean(data)
			std = np.std(data)
			print ("\t\tSample size\t", len(data))
			print ("\t\tSample mean\t", mean)
			print ("\t\tSample stdDev\t", std)
			if std > 0:
				acf = autocorr_func(data)
				acf_a = autocorr_analysis(acf)
				print ("\t\tACF\t", acf_a)
				neff_int = len(data) / acf_a.ac_time_int
				print ("\t\tEffective sample size integrated\t", neff_int)
				neff_exp = len(data) / (2. * acf_a.ac_time_exp)
				print ("\t\tEffective sample size exponential\t", neff_exp)
				if abs(1. - neff_exp/neff_int) > 0.1:
					print ("\t\t\033[1;33mAchtung!\033[0;0m Integrated and exponential autocorrelation times too different! Probably non-equilibrium state")
				print ("\t\tSample mean stdDev\t", std / neff_int)
			
		data = m.full_params_log[Ti,:,param_indices]
		
		corrs = np.corrcoef(data, rowvar=True)
		print ("\tCorrelations between parameters:")
		for row in corrs:
			print ('\t\t', ' '.join(map("{:.3f}".format, row)))
		
		
	
	
